@@ -889,6 +889,47 @@ g_usb_device_get_string_descriptor (GUsbDevice  *device,
 	return g_strdup ((const gchar *)buf);
 }
 
+/**
+ * g_usb_device_get_string_descriptor_bytes:
+ * @desc_index: the index for the string descriptor to retrieve
+ * @langid: the language ID
+ * @error: a #GError, or %NULL
+ *
+ * Get a raw string descriptor from the device. The returned string should be freed
+ * with g_bytes_unref() when no longer needed.
+ *
+ * Return value: (transfer full): a possibly UTF-16 string, or NULL on error.
+ *
+ * Since: 0.3.6
+ **/
+GBytes *
+g_usb_device_get_string_descriptor_bytes (GUsbDevice  *device,
+					  guint8       desc_index,
+					  guint16      langid,
+					  GError     **error)
+{
+	gint rc;
+	unsigned char buf[128];
+
+	g_return_val_if_fail (G_USB_IS_DEVICE (device), NULL);
+	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+	if (device->priv->handle == NULL) {
+		g_usb_device_not_open_error (device, error);
+		return NULL;
+	}
+
+	rc = libusb_get_string_descriptor (device->priv->handle,
+					   desc_index, langid,
+					   buf, sizeof(buf));
+	if (rc < 0) {
+		g_usb_device_libusb_error_to_gerror (device, rc, error);
+		return NULL;
+	}
+
+	return g_bytes_new (buf, rc);
+}
+
 typedef gssize (GUsbDeviceTransferFinishFunc) (GUsbDevice *device, GAsyncResult *res, GError **error);
 
 typedef struct {
@@ -919,7 +960,7 @@ g_usb_device_sync_transfer_cb (GUsbDevice     *device,
  * either input or output
  * @length: the length field for the setup packet.
  * @actual_length: (out) (optional): the actual number of bytes sent, or %NULL
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -986,7 +1027,7 @@ g_usb_device_control_transfer (GUsbDevice	    *device,
  * either input or output
  * @length: the length field for the setup packet.
  * @actual_length: (out) (optional): the actual number of bytes sent, or %NULL
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -1043,7 +1084,7 @@ g_usb_device_bulk_transfer (GUsbDevice    *device,
  * either input or output
  * @length: the length field for the setup packet.
  * @actual_length: (out) (optional): the actual number of bytes sent, or %NULL
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -1223,7 +1264,7 @@ g_usb_device_control_transfer_cb (struct libusb_transfer *transfer)
  * @data: (array length=length): a suitably-sized data buffer for
  * either input or output
  * @length: the length field for the setup packet.
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -1267,18 +1308,14 @@ g_usb_device_control_transfer_async (GUsbDevice	   *device,
 	req->transfer = libusb_alloc_transfer (0);
 	req->data = data;
 
-	/* setup cancellation */
-	if (cancellable != NULL) {
-		req->cancellable = g_object_ref (cancellable);
-		req->cancellable_id = g_cancellable_connect (req->cancellable,
-							     G_CALLBACK (g_usb_device_cancelled_cb),
-							     req,
-							     NULL);
-	}
-
 	task = g_task_new (device, cancellable, callback, user_data);
 	g_task_set_task_data (task, req, (GDestroyNotify)g_usb_device_req_free);
 
+	if (g_task_return_error_if_cancelled (task)) {
+		g_object_unref (task);
+		return;
+	}
+
 	/* munge back to flags */
 	if (direction == G_USB_DEVICE_DIRECTION_DEVICE_TO_HOST)
 		request_type_raw |= 0x80;
@@ -1307,6 +1344,15 @@ g_usb_device_control_transfer_async (GUsbDevice	   *device,
 		g_task_return_error (task, error);
 		g_object_unref (task);
 	}
+
+	/* setup cancellation after submission */
+	if (cancellable != NULL) {
+		req->cancellable = g_object_ref (cancellable);
+		req->cancellable_id = g_cancellable_connect (req->cancellable,
+							     G_CALLBACK (g_usb_device_cancelled_cb),
+							     req,
+							     NULL);
+	}
 }
 
 /**
@@ -1340,7 +1386,7 @@ g_usb_device_control_transfer_finish (GUsbDevice    *device,
  * @data: (array length=length): a suitably-sized data buffer for
  * either input or output
  * @length: the length field for the setup packet.
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -1377,18 +1423,14 @@ g_usb_device_bulk_transfer_async (GUsbDevice	  *device,
 	req = g_slice_new0 (GcmDeviceReq);
 	req->transfer = libusb_alloc_transfer (0);
 
-	/* setup cancellation */
-	if (cancellable != NULL) {
-		req->cancellable = g_object_ref (cancellable);
-		req->cancellable_id = g_cancellable_connect (req->cancellable,
-							     G_CALLBACK (g_usb_device_cancelled_cb),
-							     req,
-							     NULL);
-	}
-
 	task = g_task_new (device, cancellable, callback, user_data);
 	g_task_set_task_data (task, req, (GDestroyNotify)g_usb_device_req_free);
 
+	if (g_task_return_error_if_cancelled (task)) {
+		g_object_unref (task);
+		return;
+	}
+
 	/* fill in transfer details */
 	libusb_fill_bulk_transfer (req->transfer,
 				   device->priv->handle,
@@ -1406,6 +1448,15 @@ g_usb_device_bulk_transfer_async (GUsbDevice	  *device,
 		g_task_return_error (task, error);
 		g_object_unref (task);
 	}
+
+	/* setup cancellation after submission */
+	if (cancellable != NULL) {
+		req->cancellable = g_object_ref (cancellable);
+		req->cancellable_id = g_cancellable_connect (req->cancellable,
+							     G_CALLBACK (g_usb_device_cancelled_cb),
+							     req,
+							     NULL);
+	}
 }
 
 /**
@@ -1439,7 +1490,7 @@ g_usb_device_bulk_transfer_finish (GUsbDevice    *device,
  * @data: (array length=length): a suitably-sized data buffer for
  * either input or output
  * @length: the length field for the setup packet.
- * @timeout: timeout timeout (in millseconds) that this function should wait
+ * @timeout: timeout timeout (in milliseconds) that this function should wait
  * before giving up due to no response being received. For an unlimited
  * timeout, use 0.
  * @cancellable: a #GCancellable, or %NULL
@@ -1476,18 +1527,14 @@ g_usb_device_interrupt_transfer_async (GUsbDevice	  *device,
 	req = g_slice_new0 (GcmDeviceReq);
 	req->transfer = libusb_alloc_transfer (0);
 
-	/* setup cancellation */
-	if (cancellable != NULL) {
-		req->cancellable = g_object_ref (cancellable);
-		req->cancellable_id = g_cancellable_connect (req->cancellable,
-							     G_CALLBACK (g_usb_device_cancelled_cb),
-							     req,
-							     NULL);
-	}
-
 	task = g_task_new (device, cancellable, callback, user_data);
 	g_task_set_task_data (task, req, (GDestroyNotify)g_usb_device_req_free);
 
+	if (g_task_return_error_if_cancelled (task)) {
+		g_object_unref (task);
+		return;
+	}
+
 	/* fill in transfer details */
 	libusb_fill_interrupt_transfer (req->transfer,
 					device->priv->handle,
@@ -1505,6 +1552,15 @@ g_usb_device_interrupt_transfer_async (GUsbDevice	  *device,
 		g_task_return_error (task, error);
 		g_object_unref (task);
 	}
+
+	/* setup cancellation after submission */
+	if (cancellable != NULL) {
+		req->cancellable = g_object_ref (cancellable);
+		req->cancellable_id = g_cancellable_connect (req->cancellable,
+							     G_CALLBACK (g_usb_device_cancelled_cb),
+							     req,
+							     NULL);
+	}
 }
 
 /**
